#ifndef DEFAULTS_H
#define DEFAULTS_H

#define DEFAULT_DEVICE "default"
#define DEFAULT_BUFFER 16

#define DEFAULT_ADDR "::"
#define DEFAULT_PORT 1350
#define DEFAULT_FRAME 960
#define DEFAULT_JITTER 16

#define DEFAULT_RATE 48000
#define DEFAULT_CHANNELS 2
#define DEFAULT_BITRATE 128

#define DEFAULT_VERBOSE 1

#endif
